/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_main.c  -- client main loop

#include "quakedef.h"

// we need to declare some mouse variables here, because the menu system
// references them even when on a unix system.

// these two are not intended to be set directly
cvar_t	cl_name = {"_cl_name", "player", true};
cvar_t	cl_color = {"_cl_color", "0", true};

cvar_t	cl_shownet = {"cl_shownet","0"};	// can be 0, 1, or 2
cvar_t	cl_nolerp = {"cl_nolerp","0"};

cvar_t	lookspring = {"lookspring","0", true};
cvar_t	lookstrafe = {"lookstrafe","0", true};
cvar_t	sensitivity = {"sensitivity","3", true};

cvar_t	m_pitch = {"m_pitch","0.022", true};
cvar_t	m_yaw = {"m_yaw","0.022", true};
cvar_t	m_forward = {"m_forward","1", true};
cvar_t	m_side = {"m_side","0.8", true};

client_static_t	cls;
client_state_t	cl;

// FIXME: put these on hunk?
efrag_t			cl_efrags[MAX_EFRAGS];
entity_t		cl_static_entities[MAX_STATIC_ENTITIES];
lightstyle_t	cl_lightstyle[MAX_LIGHTSTYLES];
dlight_t		cl_dlights[MAX_DLIGHTS];

int				cl_numvisedicts;
entity_t		*cl_visedicts[MAX_VISEDICTS];

// These are now on hunk
int				cl_max_edicts;
entity_t		*cl_entities;

/*
=====================
CL_ClearState
=====================
*/
void CL_ClearState (void)
{
	int			i;

	if (!sv.active)
	{
		Host_ClearMemory ();
	}

	// wipe the entire cl structure
	memset (&cl, 0, sizeof(cl));

	SZ_Clear (&cls.message);

	// clear other arrays	
	memset (cl_efrags, 0, sizeof(cl_efrags));
	memset (cl_dlights, 0, sizeof(cl_dlights));
	memset (cl_lightstyle, 0, sizeof(cl_lightstyle));
	memset (cl_temp_entities, 0, sizeof(cl_temp_entities));
	memset (cl_beams, 0, sizeof(cl_beams));

	// now dynamically allocated
	cl_max_edicts = bound (256, (int)max_edicts.value, 8192);
	cl_entities = Heap_ZAlloc (cl_max_edicts * sizeof(entity_t));

	// allocate the efrags and chain together into a free list
	cl.free_efrags = cl_efrags;

	for (i=0 ; i<MAX_EFRAGS-1 ; i++)
	{
		cl.free_efrags[i].entnext = &cl.free_efrags[i+1];
	}
	cl.free_efrags[i].entnext = NULL;
}

/*
=====================
CL_Disconnect

Sends a disconnect message to the server
This is also called on Host_Error, so it shouldn't cause any errors
=====================
*/
void CL_Disconnect (void)
{
	// stop sounds (especially looping!)
	S_StopAllSounds (true);
	
	// if running a local server, shut it down
	if (cls.demoplayback)
	{
		CL_StopPlayback ();
	}
	else if (cls.state == ca_connected)
	{
		if (cls.demorecording)
		{
			CL_Stop_f ();
		}
		Con_DPrintf ("Sending clc_disconnect\n");
		SZ_Clear (&cls.message);
		MSG_WriteByte (&cls.message, clc_disconnect);
		NET_SendUnreliableMessage (cls.netcon, &cls.message);
		SZ_Clear (&cls.message);
		NET_Close (cls.netcon);

		cls.state = ca_disconnected;

		if (sv.active)
		{
			Host_ShutdownServer(false);
		}
	}
	cls.demoplayback = cls.timedemo = false;
	cls.signon = 0;
}

/*
=====================
CL_Disconnect_f
=====================
*/
void CL_Disconnect_f (void)
{
	CL_Disconnect ();

	if (sv.active)
	{
		Host_ShutdownServer (false);
	}
}

/*
=====================
CL_EstablishConnection

Host should be either "local" or a net address to be passed on
=====================
*/
void CL_EstablishConnection (char *host)
{
	if (cls.state == ca_dedicated) return;
	if (cls.demoplayback) return;

	CL_Disconnect ();

	cls.netcon = NET_Connect (host);

	if (!cls.netcon)
	{
		Host_Error ("CL_Connect: connect failed\n");
	}
	Con_DPrintf ("CL_EstablishConnection: connected to %s\n", host);
	
	cls.demonum = -1;			// not in the demo loop now
	cls.state = ca_connected;
	cls.signon = 0;				// need all the signon messages before playing
}

/*
=====================
CL_SignonReply

An svc_signonnum has been received, perform a client side setup
=====================
*/
void CL_SignonReply (void)
{
	Con_DPrintf ("CL_SignonReply: %i\n", cls.signon);

	switch (cls.signon)
	{
	case 1:
		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, "prespawn");
		break;
		
	case 2:		
		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, va("name \"%s\"\n", cl_name.string));
	
		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, va("color %i %i\n", ((int)cl_color.value)>>4, ((int)cl_color.value)&15));
	
		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, va("spawn %s", cls.spawnparms));
		break;
		
	case 3:	
		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, "begin");
		break;
		
	case 4:
		SCR_EndLoadingPlaque ();		// allow normal screen updates
		break;
	}
}

/*
=====================
CL_NextDemo

Called to play the next demo in the demo loop
=====================
*/
void CL_NextDemo (void)
{
	if (cls.demonum == -1) return;		// don't play demos

	SCR_BeginLoadingPlaque ();

	if (!cls.demos[cls.demonum][0] || cls.demonum == MAX_DEMOS)
	{
		cls.demonum = 0;

		if (!cls.demos[cls.demonum][0])
		{
			Con_Printf ("No demos listed with startdemos\n");
			cls.demonum = -1;
			return;
		}
	}
	Cbuf_InsertText (va("playdemo %s\n", cls.demos[cls.demonum]));
	cls.demonum++;
}

/*
==============
CL_PrintEntities_f
==============
*/
void CL_PrintEntities_f (void)
{
	entity_t	*ent;
	int			i;
	
	for (i=0,ent=cl_entities ; i<cl.num_entities ; i++,ent++)
	{
		Con_Printf ("%3i:",i);

		if (!ent->model)
		{
			Con_Printf ("EMPTY\n");
			continue;
		}
		Con_Printf ("%s:%2i  (%5.1f,%5.1f,%5.1f) [%5.1f %5.1f %5.1f]\n"
		,ent->model->name,ent->frame, ent->origin[0], ent->origin[1], ent->origin[2], ent->angles[0], ent->angles[1], ent->angles[2]);
	}
}

/*
===============
CL_AllocDlight

===============
*/
dlight_t *CL_AllocDlight (int key)
{
	int			i;
	dlight_t	*dl;

	// first look for an exact key match
	if (key)
	{
		dl = cl_dlights;

		for (i=0 ; i<MAX_DLIGHTS ; i++, dl++)
		{
			if (dl->key == key)
			{
				memset (dl, 0, sizeof(*dl));
				dl->key = key;

				dl->RGB[0] = 1;
				dl->RGB[1] = 1;
				dl->RGB[2] = 1;

				return dl;
			}
		}
	}

	// then look for anything else
	dl = cl_dlights;

	for (i=0 ; i<MAX_DLIGHTS ; i++, dl++)
	{
		if (dl->die < cl.time)
		{
			memset (dl, 0, sizeof(*dl));
			dl->key = key;
			dl->RGB[0] = 1;
			dl->RGB[1] = 1;
			dl->RGB[2] = 1;
			return dl;
		}
	}

	dl = &cl_dlights[0];

	memset (dl, 0, sizeof(*dl));
	dl->RGB[0] = 1;
	dl->RGB[1] = 1;
	dl->RGB[2] = 1;
	dl->key = key;

	return dl;
}


/*
===============
CL_DecayLights

===============
*/
void CL_DecayLights (void)
{
	int			i;
	dlight_t	*dl;
	float		time;
	
	time = cl.time - cl.oldtime;

	dl = cl_dlights;

	for (i=0 ; i<MAX_DLIGHTS ; i++, dl++)
	{
		if (dl->die < cl.time || !dl->radius) continue;
		
		dl->radius -= time*dl->decay;

		if (dl->radius < 0)
		{
			dl->radius = 0;
		}
	}
}


/*
===============
CL_LerpPoint

Determines the fraction between the last two messages that the objects
should be put at.
===============
*/
float	CL_LerpPoint (void)
{
	float	f, frac;

	f = cl.mtime[0] - cl.mtime[1];
	
	if (!f || cl_nolerp.value || cls.timedemo || sv.active)
	{
		cl.time = cl.mtime[0];
		return 1;
	}
		
	if (f > 0.1)
	{	
		// dropped packet, or start of demo
		cl.mtime[1] = cl.mtime[0] - 0.1;
		f = 0.1f;
	}
	frac = (cl.time - cl.mtime[1]) / f;

	if (frac < 0)
	{
		if (frac < -0.01)
		{
			cl.time = cl.mtime[1];
		}
		frac = 0;
	}
	else if (frac > 1)
	{
		if (frac > 1.01)
		{
			cl.time = cl.mtime[0];
		}
		frac = 1;
	}

	if (cl_nolerp.value)
	{
		return 1;
	}
	return frac;
}


/*
===============
CL_RelinkEntities
===============
*/
void CL_RelinkEntities (void)
{
	entity_t	*ent;
	int			i, j;
	float		frac, f, d;
	vec3_t		delta;
	float		objrotate;
	vec3_t		oldorg;
	float		lastmsg ;
	dlight_t	*dl;

	// terminate lastmessage properly.
	memset (&lastmsg, 0, sizeof(float));

	// determine partial update time
	frac = CL_LerpPoint ();

	cl_numvisedicts = 0;

	// interpolate player info
	for (i = 0; i < 3; i++)
	{
		cl.velocity[i] = cl.mvelocity[1][i] + frac * (cl.mvelocity[0][i] - cl.mvelocity[1][i]);
	}

	if (cls.demoplayback)
	{
		// interpolate the angles	
		for (j=0 ; j<3 ; j++)
		{
			d = cl.mviewangles[0][j] - cl.mviewangles[1][j];

			if (d > 180)
			{
				d -= 360;
			}
			else if (d < -180)
			{
				d += 360;
			}
			cl.viewangles[j] = cl.mviewangles[1][j] + frac*d;
		}
	}
	objrotate = anglemod(100*cl.time);

	// start on the entity after the world
	for (i=1,ent=cl_entities+1 ; i<cl.num_entities ; i++,ent++)
	{
		if (!ent->model)
		{
			// empty slot
			if (ent->forcelink)
			{
				R_RemoveEfrags (ent);	// just became empty
			}
			continue;
		}

		// if the object wasn't included in the last packet, remove it
		if (ent->msgtime != cl.mtime[0])
		{
			ent->model = NULL;

			ent->frameStartTime = 0;
			ent->translateStartTime = 0;
			ent->rotateStartTime = 0;

			continue;
		}
		VectorCopy (ent->origin, oldorg);

		if (ent->forcelink)
		{
			// the entity was not updated in the last message
			// so move to the final spot
			VectorCopy (ent->msg_origins[0], ent->origin);
			VectorCopy (ent->msg_angles[0], ent->angles);
		}
		else
		{
			// if the delta is large, assume a teleport and don't lerp
			f = frac;

			for (j=0 ; j<3 ; j++)
			{
				delta[j] = ent->msg_origins[0][j] - ent->msg_origins[1][j];

				if (delta[j] > 100 || delta[j] < -100)
				{
					f = 1;		// assume a teleportation, not a motion
				}
			}

			if (f >= 1)
			{
				ent->frameStartTime = 0;
				ent->translateStartTime = 0;
				ent->rotateStartTime = 0;
			}

			// interpolate the origin and angles
			for (j=0 ; j<3 ; j++)
			{
				ent->origin[j] = ent->msg_origins[1][j] + f*delta[j];

				d = ent->msg_angles[0][j] - ent->msg_angles[1][j];

				if (d > 180)
				{
					d -= 360;
				}
				else if (d < -180)
				{
					d += 360;
				}
				ent->angles[j] = ent->msg_angles[1][j] + f*d;
			}
			
		}

		// rotate binary objects locally
		if (ent->model->flags & EF_ROTATE)
		{
			ent->angles[1] = objrotate;
		}

		if (ent->effects & EF_BRIGHTFIELD)
		{
			R_EntityParticles (ent);
		}

		if (ent->effects & EF_MUZZLEFLASH)
		{
			vec3_t		fv, rv, uv;
			vec3_t		smokeorg;

			dl = CL_AllocDlight (i);
			VectorCopy (ent->origin,  dl->origin);
			dl->origin[2] += 18;

			AngleVectors (ent->angles, fv, rv, uv);
			 
			VectorMA (dl->origin, 18, fv, dl->origin);
			dl->radius = 200 + (Q_rand ()&31);
			dl->minlight = 32;
			dl->die = cl.time + 0.1;

			// base RGB
			dl->RGB[0] = 1;
			dl->RGB[1] = 0.5;
			dl->RGB[2] = 0;

			// switch the dlight colour for the wizard
			if (ent->model->name[6] == 'w' &&
				ent->model->name[7] == 'i' &&
				ent->model->name[8] == 'z' &&
				ent->model->name[8] == 'a' &&
				ent->model->name[9] == 'r' &&
				ent->model->name[10] == 'd')
			{
				// different muzzleflash colour for wizards
				dl->RGB[0] = 0.1f;
				dl->RGB[1] = 0.5f;
				dl->RGB[2] = 0.1f;
			}
			else if (ent->model->name[6] == 's' &&
				ent->model->name[7] == 'h' &&
				ent->model->name[8] == 'a' &&
				ent->model->name[8] == 'l' &&
				ent->model->name[9] == 'r' &&
				ent->model->name[10] == 'a' &&
				ent->model->name[11] == 't' &&
				ent->model->name[12] == 'h')
			{
				// different muzzleflash colour for vores
				dl->RGB[0] = 0.3f;
				dl->RGB[1] = 0.1f;
				dl->RGB[2] = 0.2f;
			}
			else if (ent->model->name[6] == 's' &&
				ent->model->name[7] == 'h' &&
				ent->model->name[8] == 'a' &&
				ent->model->name[8] == 'm' &&
				ent->model->name[9] == 'b' &&
				ent->model->name[10] == 'l' &&
				ent->model->name[11] == 'e' &&
				ent->model->name[12] == 'r')
			{
				// different muzzleflash colour for shamblers
				dl->RGB[0] = 0.02f;
				dl->RGB[1] = 0.2f;
				dl->RGB[2] = 0.5f;
			}

			// check for progs/player.mdl and if cl.stats[STAT_ACTIVEWEAPON] == whatever
			// and change the muzzle flash colour depending on the weapon
			if (i == cl.viewentity)
			{
				if (cl.stats[STAT_ACTIVEWEAPON] == IT_LIGHTNING)
				{
					dl->RGB[0] = 0.1f;
					dl->RGB[1] = 0.1f;
					dl->RGB[2] = 0.5f;
				}

				// smoking pistols!!!  yee-haw
				// to do - add rogue and hipnotic too...
				VectorCopy (ent->origin, smokeorg);

				smokeorg[2] += 15;	// move it up

				// adjust for current view angles
				AngleVectors (cl.viewangles, fv, rv, uv);

				// forward vector
				VectorMA (smokeorg, 18, fv, smokeorg);

				if (cl.stats[STAT_ACTIVEWEAPON] == IT_NAILGUN || cl.stats[STAT_ACTIVEWEAPON] == IT_SUPER_LIGHTNING)
				{
					// because these db weapons generate 2 smoke effects, reduce the count to compensate
					VectorMA (smokeorg, -3, rv, smokeorg);
					R_SmokeEffect (smokeorg, 3, 1.0);

					VectorMA (smokeorg, 6, rv, smokeorg);
					R_SmokeEffect (smokeorg, 3, 1.0);
				}
				else if (cl.stats[STAT_ACTIVEWEAPON] == IT_SUPER_NAILGUN || cl.stats[STAT_ACTIVEWEAPON] == IT_LIGHTNING)
				{
					// rapid fires have less smoke
					R_SmokeEffect (smokeorg, 4, 1.0);
				}
				else
				{
					// 8 is the norm for all other weapons
					R_SmokeEffect (smokeorg, 8, 1.0);
				}
			}
		}

		if (ent->effects & EF_BRIGHTLIGHT)
		{			
			dl = CL_AllocDlight (i);
			VectorCopy (ent->origin,  dl->origin);
			dl->origin[2] += 16;
			dl->radius = 400 + (Q_rand ()&31);
			dl->die = cl.time + 0.001;
		}
		if (ent->effects & EF_DIMLIGHT)
		{			
			dl = CL_AllocDlight (i);
			VectorCopy (ent->origin,  dl->origin);
			dl->radius = 200 + (Q_rand ()&31);
			dl->die = cl.time + 0.001;

			// powerup dynamic lights
			if (i == cl.viewentity)
			{
				dl->RGB[0] = 0.1f;
				dl->RGB[1] = 0.1f;
				dl->RGB[2] = 0.1f;

				if (cl.items & IT_QUAD) dl->RGB[2] = 0.5;
				if (cl.items & IT_INVULNERABILITY) dl->RGB[0] = 0.5;
			}
		}

		// if model has no flags dont check each
		if (ent->model->flags) 
		{
			if (ent->model->flags & EF_GIB)
			{
				// blood splash
				// if all 3 deltas are non-zero assume a moving gib (some come in from the 
				// server with delta[2] == 16 on a new visframe after being spawned at 
				// their origin then moving to the floor)
				if (delta[0] && delta[1] && delta[2] && delta[2] != 16)
				{
					R_AddStain (ent->origin, 0.50, 0.125, 0.0625, Q_rand () & 25);

					// don't do the rockettrail if there's no movement in the gib either
					R_RocketTrail (oldorg, ent->origin, 2);
				}
			}
			else if (ent->model->flags & EF_ZOMGIB)
			{
				// these are the opposite - they only add a stain if they're not
				// moving cos they disappear immediately after!!!
				if (!delta[0] && !delta[1] && !delta[2])
				{
					R_AddStain (ent->origin, 0.50, 0.125, 0.0625, Q_rand () & 25);
				}
				R_RocketTrail (oldorg, ent->origin, 4);
			}
			else if (ent->model->flags & EF_TRACER)
			{
				// wizard
				R_RocketTrail (oldorg, ent->origin, 3);

				dl = CL_AllocDlight (i);
				VectorCopy (ent->origin, dl->origin);
				dl->radius = 200;
				dl->die = cl.time + 0.01;
				dl->RGB[0] = 0.1f;
				dl->RGB[1] = 0.5f;
				dl->RGB[2] = 0.1f;

				dl->decay = 100;
			}
			else if (ent->model->flags & EF_TRACER2)
			{
				// hellknight
				R_RocketTrail (oldorg, ent->origin, 5);

				dl = CL_AllocDlight (i);
				VectorCopy (ent->origin, dl->origin);
				dl->radius = 200;
				dl->die = cl.time + 0.01;

				dl->decay = 100;
			}
			else if (ent->model->flags & EF_ROCKET)
			{
				R_RocketTrail (oldorg, ent->origin, 0);
				dl = CL_AllocDlight (i);
				VectorCopy (ent->origin, dl->origin);
				dl->radius = 200;
				dl->die = cl.time + 0.01;
			}
			else if (ent->model->flags & EF_GRENADE)
			{
				R_RocketTrail (oldorg, ent->origin, 1);
			}
			else if (ent->model->flags & EF_TRACER3)
			{
				// vore
				R_RocketTrail (oldorg, ent->origin, 6);

				dl = CL_AllocDlight (i);
				VectorCopy (ent->origin, dl->origin);
				dl->radius = 200;
				dl->die = cl.time + 0.01;
				dl->RGB[0] = 0.5f;
				dl->RGB[1] = 0.1f;
				dl->RGB[2] = 0.5f;

				dl->decay = 100;
			}
		}
		ent->forcelink = false;

		if (i == cl.viewentity && !chase_active.value) continue;

		if (cl_numvisedicts < MAX_EDICTS)
		{
			cl_visedicts[cl_numvisedicts++] = ent;
		}
		else if (COM_Timeout (&lastmsg, 2))
		{
			Con_DPrintf ("CL_RelinkEntities: too many visedicts (max = %d)\n", cl_max_edicts);
		}
	}
}


/*
===============
CL_ReadFromServer

Read all incoming data from the server
===============
*/
int CL_ReadFromServer (void)
{
	int				ret;
	static qboolean WarnPrinted;

	cl.oldtime = cl.time;
	cl.time += host_frametime;
	
	do
	{
		ret = CL_GetMessage ();

		if (ret == -1)
		{
			Host_Error ("CL_ReadFromServer: lost server connection");
		}

		if (!ret) break;		
		cl.last_received_message = realtime;
		CL_ParseServerMessage ();
	} while (ret && cls.state == ca_connected);
	
	if (cl_shownet.value)
	{
		Con_Printf ("\n");
	}

	if (cl.time < 1)
	{
		// Kludge to reset at beginning of new map
		if (cl.num_statics <= 128)
		{
			WarnPrinted = false;
		}
	}
	else
	{
		// Check old limit
		if (!WarnPrinted && cl.num_statics > 128)
		{
			WarnPrinted = true;
			Con_Printf ("\002CL_ReadFromServer: ");
			Con_Printf ("excessive static entities (%d, normal max = %d)\n", cl.num_statics, 128);
		}
	}
	CL_RelinkEntities ();
	CL_UpdateTEnts ();
//
// bring the links up to date
//
	return 0;
}

/*
=================
CL_SendCmd
=================
*/
void CL_SendCmd (void)
{
	usercmd_t		cmd;

	if (cls.state != ca_connected) return;

	if (cls.signon == SIGNONS)
	{
	// get basic movement from keyboard
		CL_BaseMove (&cmd);
	
	// allow mice or other external controllers to add to the move
		IN_Move (&cmd);
	
	// send the unreliable message
		CL_SendMove (&cmd);
	
	}

	if (cls.demoplayback)
	{
		SZ_Clear (&cls.message);
		return;
	}
	
// send the reliable message
	if (!cls.message.cursize) return;		// no message at all
	
	if (!NET_CanSendMessage (cls.netcon))
	{
		Con_DPrintf ("CL_WriteToServer: can't send\n");
		return;
	}

	if (NET_SendMessage (cls.netcon, &cls.message) == -1)
	{
		Host_Error ("CL_WriteToServer: lost server connection");
	}
	SZ_Clear (&cls.message);
}

/*
=============
CL_Viewpos_f

Display client's position and mangle
=============
*/
void CL_Viewpos_f (void)
{
	// Player position
	Con_Printf ("Viewpos: (%i %i %i) %i %i %i\n",
		(int)cl_entities[cl.viewentity].origin[0],
		(int)cl_entities[cl.viewentity].origin[1],
		(int)cl_entities[cl.viewentity].origin[2],
		(int)cl.viewangles[PITCH],
		(int)cl.viewangles[YAW],
		(int)cl.viewangles[ROLL]);
		
	if (sv_player && sv_player->v.movetype == MOVETYPE_NOCLIP && Cmd_Argc() == 4)
	{
		// Major hack ...
		sv_player->v.origin[0] = atoi (Cmd_Argv(1));
		sv_player->v.origin[1] = atoi (Cmd_Argv(2));
		sv_player->v.origin[2] = atoi (Cmd_Argv(3));
	}
}

/*
=============
CL_Mapname_f
=============
*/
void CL_Mapname_f (void)
{
	char name[MAX_QPATH];
	
	name[0] = '\0';
	
	if (cl.worldmodel && cl.worldmodel->name)
	{
		COM_StripExtension (cl.worldmodel->name + 5, name);
	}
	Con_Printf ("\"mapname\" is \"%s\"\n", name);
}

/*
=============
CL_StaticEnts_f
=============
*/
void CL_StaticEnts_f (void)
{
	Con_Printf ("%d static entities\n", cl.num_statics);
}

/*
=================
CL_Init
=================
*/
void CL_Init (void)
{	
	SZ_Alloc (&cls.message, 8192);
	cls.message.function = "CL_Init";

	CL_InitInput ();
	CL_InitTEnts ();
	
//
// register our commands
//
	Cvar_RegisterVariable (&cl_name);
	Cvar_RegisterVariable (&cl_color);
	Cvar_RegisterVariable (&cl_upspeed);
	Cvar_RegisterVariable (&cl_forwardspeed);
	Cvar_RegisterVariable (&cl_backspeed);
	Cvar_RegisterVariable (&cl_sidespeed);
	Cvar_RegisterVariable (&cl_movespeedkey);
	Cvar_RegisterVariable (&cl_yawspeed);
	Cvar_RegisterVariable (&cl_pitchspeed);
	Cvar_RegisterVariable (&cl_anglespeedkey);
	Cvar_RegisterVariable (&cl_shownet);
	Cvar_RegisterVariable (&cl_nolerp);
	Cvar_RegisterVariable (&lookspring);
	Cvar_RegisterVariable (&lookstrafe);
	Cvar_RegisterVariable (&sensitivity);

	Cvar_RegisterVariable (&m_pitch);
	Cvar_RegisterVariable (&m_yaw);
	Cvar_RegisterVariable (&m_forward);
	Cvar_RegisterVariable (&m_side);

	Cmd_AddCommand ("entities", CL_PrintEntities_f);
	Cmd_AddCommand ("disconnect", CL_Disconnect_f);
	Cmd_AddCommand ("record", CL_Record_f);
	Cmd_AddCommand ("stop", CL_Stop_f);
	Cmd_AddCommand ("playdemo", CL_PlayDemo_f);
	Cmd_AddCommand ("timedemo", CL_TimeDemo_f);
	Cmd_AddCommand ("viewpos", CL_Viewpos_f);
	Cmd_AddCommand ("mapname", CL_Mapname_f);
	Cmd_AddCommand ("staticents", CL_StaticEnts_f);
}

